﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.ParameterConfigurationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// TODO
  /// </summary>
  [Item("SolutionCacheAnalyzer", "TODO")]
  [StorableClass]
  public sealed class SolutionCacheAnalyzer : SingleSuccessorOperator, IAnalyzer {
    public bool EnabledByDefault {
      get { return true; }
    }

    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ScopeTreeLookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ScopeTreeLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ValueParameter<BoolValue> StoreAllRunsParameter {
      get { return (ValueParameter<BoolValue>)Parameters["StoreAllRuns"]; }
    }

    #region Constructors and Cloning
    public SolutionCacheAnalyzer()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", ""));
      Parameters.Add(new ScopeTreeLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new ValueParameter<BoolValue>("StoreAllRuns", "If true all runs ever executed are stored. Otherwise only the runs from the latest generateion are stored for caching purposes.", new BoolValue(false)));
    }

    [StorableConstructor]
    private SolutionCacheAnalyzer(bool deserializing) : base(deserializing) { }
    private SolutionCacheAnalyzer(SolutionCacheAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SolutionCacheAnalyzer(this, cloner);
    }
    #endregion

    public override IOperation Apply() {
      ResultCollection results = ResultsParameter.ActualValue;
      ItemArray<ParameterConfigurationTree> solutions = ParameterConfigurationParameter.ActualValue;
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      ItemDictionary<StringValue, RunCollection> allRuns = ResultsParameter.ActualValue.ContainsKey("SolutionCache") ? (ItemDictionary<StringValue, RunCollection>)ResultsParameter.ActualValue["SolutionCache"].Value : new ItemDictionary<StringValue, RunCollection>();
      bool storeAllRuns = ((BoolValue)StoreAllRunsParameter.ActualValue).Value;

      if (!storeAllRuns) {
        allRuns.Clear();
      }

      foreach (var solution in solutions) {
        string key = solution.ParameterInfoString;
        bool first = false;
        if (allRuns.Count(x => x.Key.Value == key) == 0) {
          allRuns.Add(new StringValue(key), new RunCollection());
          first = true; // no other runs yet
        }

        if (solution.Runs != null) { // Runs is null when a base-level algorithm exception happened due to invalid parameters
          var runCollection = allRuns.Single(x => x.Key.Value == key).Value;
          foreach (var run in solution.Runs) {
            if (!((BoolValue)run.Results["Meta-FromCache"]).Value || first) {
              run.Results["Meta-FromCache"] = new BoolValue(true);
              runCollection.Add(run);
            }
          }
        }
      }

      if (!results.ContainsKey("SolutionCache")) {
        results.Add(new Result("SolutionCache", allRuns));
      }

      return base.Apply();
    }
  }
}
